/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.reader;

import cricketdaemon.CricketDaemon;
import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.beacontable.*;
import cricketdaemon.reader.parse.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Retrieves and parses beacon data from the gpsd.
 *
 * @author Allen Miu
 */
public class GPSDReader extends ModuleThread
{
    private BufferedReader in = null;
    private OutputStream out = null;

    private final String REGISTRATION_COMMAND = "r\n";
    private final String ENCODING = "UTF-8";
    private final String TOKEN_DELIM = ",";
    private final String VALUE_DELIM = "=";

    protected String MAGIC_TOKEN = "$Cricket2";
    protected String addr;
    protected int port;

    public GPSDReader(String addr, int port)
    {
	super("GPSDReader", 0);
	this.addr = addr;
	this.port = port;
	connect();
    }

    protected synchronized void connect()
    {
	while(true) {
	    try {
		if(in != null)
		    in.close();
		if(out != null)
		    out.close();

		System.out.println("Connecting to cricketd at "+addr);
		
		Socket s = new Socket(InetAddress.getByName(addr), port);
		InputStreamReader isr  = 
		    new InputStreamReader(s.getInputStream(), ENCODING);
		in  = new BufferedReader(isr);
		out = s.getOutputStream();
		
		sendRegistration(out);

		System.out.println("Connected to cricketd");

		return; // success
	    }
	    catch (UnknownHostException e) {
		Util.warn("GPSDReader() invalid cricketd address "+addr);
	    }
	    catch (IOException e) {
		Util.warn("GPSDReader() error contacting cricketd at "+
			  addr+" port "+port);
	    }

	    try {
		wait(500);
	    }
	    catch(InterruptedException ie) {
	    }
	}
    }

    private void sendRegistration(OutputStream out) 
	throws IOException
    {
	try {
	    byte[] r = REGISTRATION_COMMAND.getBytes(ENCODING);
	    out.write(r);
	}
	catch (UnsupportedEncodingException e) {
	    Util.assert_nonjava(false, "GPSDReader::sendRegistration() system does "+
			"not support "+ENCODING+" encoding");
	}
    }

    public void run()
    {
	while(true)
	    process();
    }

    protected void moduleCleanUp()
    {  // never really gets called because run() is overridden
	return;
    }

    protected boolean insertVal(Beacon beacon, BitSet mask, String t) {
	int i = t.indexOf(VALUE_DELIM);
	if(i <= 0) {
	    Util.log("GPSDReader::process() invalid token: "+t);
	    return false; // silently drops
	}
	String type  = t.substring(0, i);
	int fieldID  = ParseMap.typeToFieldID(type);
	
	if(fieldID < 0) {
	    Util.log("GPSDReader::process() invalid type: "+t);
	    return false; // silently drops
	}

	// ignore "null" value field values
	if(t.length() > i+1) {
	    String value = t.substring(i+1, t.length());
	    FieldParser f = ParseMap.map[fieldID];
	    Object o = f.GPSDStringToReturnType(value);
	    if(o == null) {
		Util.log("GPSDReader::process() invalid value: "+t);
		return false;
	    }
	    beacon.insertValue(f.unparse(o), fieldID);
	    mask.set(fieldID);
	}
	
	return true;
    }


    private int ct = 0;

    protected void process()
    {
	String inputStr = null;
	try {
	    //System.out.println("GPSD: readLine");
	    inputStr = in.readLine();
	    //System.out.println("GPSD: line read");

	    if(CricketDaemon.rawGPSDPrint) {
		System.out.println("GPSD "+ct+": "+inputStr);
		ct++;
	    }

	    if(inputStr == null) {
		throw new IOException("GPSD returns null");
	    }

	    StringTokenizer s = 
		new StringTokenizer(inputStr, TOKEN_DELIM);

	    boolean validDistance = false;

	    if(s.hasMoreTokens() && MAGIC_TOKEN.equals(s.nextToken())) {
		Beacon beacon = new Beacon(System.currentTimeMillis());
		BitSet mask   = (BitSet) CLEARMASK.clone();
		while(s.hasMoreTokens()) {
		    String t = s.nextToken();
		    if(insertVal(beacon, mask, t))
			validDistance = true;
		}

		/* GPSD v3 build 1 sometimes omit the distance field
		   if it only hears the radio but not the
		   ultrasound.  Here, we're filtering out all
		   readings without a distance because some
		   CricketDaemon may assume the distance field is
		   always available.  */
		// FIXME:  should eventually remove this filter
		if(mask.get(ParseMap.DISTANCE))
		    transfer(beacon, mask);
		// else silently drop the reading
	    }
	}
	catch (IOException e) {
	    //Util.assert_nonjava(false, "GPSDReader::process() read error");
	    Util.warn("GPSDReader::process() read error: "+inputStr);
	    connect();
	}
    }

    public static void main(String[] args) 
    {
	try {
	    CricketDaemon.rawGPSDPrint = true;
	    GPSDReader g = 
		new GPSDReader("18.24.6.238", 2947);
	    DebugModule d = new DebugModule(g);

	    /* comment out the above and uncomment this for translation
	       GPSDReader g = 
	       new GPSDReader(InetAddress.getByName("18.24.6.238"), 2947);
	       BeaconTableModule b = new BeaconTableModule(g, "translate.conf");
	       DebugModule d = new DebugModule(b);
	       b.start();
	    */ 

	    g.start();
	    d.start();
	    g.join();
	}
	catch (InterruptedException e) {
	}
    }
}

